@@ -116,10 +116,15 @@ static void scrolledwin_size_allocate(GtkWidget *widget,
 				      GtkAllocation *allocation,
 				      gpointer userdata);
 
+static void scrolledwin2_size_request(GtkWidget *widget,
+				      GtkRequisition *requisition,
+				      gpointer userdata);
+
 static gboolean mainview_auto_connect(gpointer userdata);
 static gboolean make_connection(const gchar *display, gboolean shared,
 				VncViewerDepthType depth,
 				VncViewerRotationType rotationType,
+				VncViewerEncodingType encodingType,
 				MainView *mainview);
 static gboolean app_key_snooper(GtkWidget *widget, GdkEventKey *keyevent,
 				HildonProgram *app);
@@ -166,6 +171,11 @@ MainView* interface_main_view_new(AppData *data)
     result->default_opts.m_nRotationType =
       (VncViewerDepthType) gconf_client_get_int(result->gc_client,
 						MAEMO_VNCVIEWER_GCONF_ROTATION, 0);
+
+    // if not found, 0 is also the default
+    result->default_opts.m_nEncodingType =
+      (VncViewerDepthType) gconf_client_get_int(result->gc_client,
+						MAEMO_VNCVIEWER_GCONF_ENCODING, 0);
     // i.e. if not found, use the default
     if (!result->default_opts.m_nDepth)
       result->default_opts.m_nDepth = VNC_VIEWER_DEPTH_DEFAULT;
@@ -293,6 +303,8 @@ void interface_main_view_destroy(MainView *mainview)
 		       (gint) mainview->default_opts.m_nDepth, 0);
   gconf_client_set_int(mainview->gc_client, MAEMO_VNCVIEWER_GCONF_ROTATION,
 		       (gint) mainview->default_opts.m_nRotationType, 0);
+  gconf_client_set_int(mainview->gc_client, MAEMO_VNCVIEWER_GCONF_ENCODING,
+		       (gint) mainview->default_opts.m_nEncodingType, 0);
 
   // using ! as per comments in interface_main_view_new()
   gconf_client_set_bool(mainview->gc_client,
@@ -342,7 +354,8 @@ void options_button_clicked(GtkButton *button,
 {
   GtkWidget *parent_dialog = GTK_WIDGET(data);
   GtkWidget *dialog, *shared_button;
-  GtkWidget *vbox, *frame;
+  GtkWidget *scrolled_window;
+  GtkWidget *vbox, *frame, *main_vbox;
   GtkToggleButton *default_radio, *eight_bit_radio, *sixteen_bit_radio;
   GtkToggleButton *toggle = 0;
 
@@ -353,25 +366,37 @@ void options_button_clicked(GtkButton *button,
 				       GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT,
 				       0);
 
-  gtk_dialog_set_has_separator(GTK_DIALOG(dialog), FALSE);
+  gtk_dialog_set_has_separator(GTK_DIALOG(dialog), TRUE);
   gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT);
 
+  scrolled_window = gtk_scrolled_window_new(0, 0);
+  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window),
+				 GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);
+  gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), scrolled_window);
+  main_vbox = gtk_vbox_new(FALSE, 0);
+  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrolled_window),
+  					main_vbox);
+  g_signal_connect(G_OBJECT(scrolled_window), "size-request",
+		   G_CALLBACK(scrolledwin2_size_request), main_vbox);
+
   options_settings *pOpts;
   pOpts = (options_settings *) g_object_get_data(G_OBJECT(parent_dialog),
 						 "options");
 
   shared_button = gtk_check_button_new_with_mnemonic("_Shared connection (do not disconnect other viewers)");
   //gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), shared_button);
-  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), shared_button,
-		     TRUE, TRUE, 5);
+  gtk_box_pack_start(GTK_BOX(main_vbox), shared_button, TRUE, TRUE, 5);
+  //gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), shared_button,
+  //	     TRUE, TRUE, 5);
   gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(shared_button),
 			       pOpts->m_bShared);
 
   frame = gtk_frame_new("Display depth");
   gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);
   //gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), frame);
-  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), frame,
-		     TRUE, TRUE, 5);
+  gtk_box_pack_start(GTK_BOX(main_vbox), frame, TRUE, TRUE, 5);
+  //gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), frame,
+  //		     TRUE, TRUE, 5);
 
   vbox = gtk_vbox_new(TRUE, 2);
   //gtk_button_box_set_layout(GTK_BUTTON_BOX(vbox), GTK_BUTTONBOX_EDGE);
@@ -408,11 +433,57 @@ void options_button_clicked(GtkButton *button,
 
   gtk_toggle_button_set_active(toggle, TRUE);
 
+  // encoding
+  GtkWidget *encoding_frame = gtk_frame_new("Preferred encoding");
+  gtk_frame_set_shadow_type(GTK_FRAME(encoding_frame), GTK_SHADOW_ETCHED_IN);
+  gtk_box_pack_start(GTK_BOX(main_vbox), encoding_frame, TRUE, TRUE, 5);
+  //gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), encoding_frame,
+  //		     TRUE, TRUE, 5);
+
+  GtkWidget *encoding_vbox = gtk_vbox_new(TRUE, 2);
+  gtk_container_add(GTK_CONTAINER(encoding_frame), encoding_vbox);
+
+  GtkToggleButton *auto_encoding_radio =
+    GTK_TOGGLE_BUTTON(gtk_radio_button_new_with_mnemonic(0, "Automatically _select"));
+  GtkToggleButton *raw_encoding_radio =
+    GTK_TOGGLE_BUTTON(gtk_radio_button_new_with_mnemonic_from_widget(GTK_RADIO_BUTTON(auto_encoding_radio), "Ra_w"));
+  GtkToggleButton *hextile_encoding_radio =
+    GTK_TOGGLE_BUTTON(gtk_radio_button_new_with_mnemonic_from_widget(GTK_RADIO_BUTTON(raw_encoding_radio), "_Hextile"));
+
+  gtk_container_add(GTK_CONTAINER(encoding_vbox),
+		    GTK_WIDGET(auto_encoding_radio));
+  gtk_container_add(GTK_CONTAINER(encoding_vbox),
+		    GTK_WIDGET(raw_encoding_radio));
+  gtk_container_add(GTK_CONTAINER(encoding_vbox),
+		    GTK_WIDGET(hextile_encoding_radio));
+
+  switch (pOpts->m_nEncodingType)
+  {
+  case VNC_VIEWER_ENCODING_AUTO:
+    toggle = auto_encoding_radio;
+    break;
+  
+  case VNC_VIEWER_ENCODING_RAW:
+    toggle = raw_encoding_radio;
+    break;
+  
+  case VNC_VIEWER_ENCODING_HEXTILE:
+    toggle = hextile_encoding_radio;
+    break;
+  
+  default:
+    break;
+  }
+
+  gtk_toggle_button_set_active(toggle, TRUE);
+
+
   // rotation
   GtkWidget *rotation_frame = gtk_frame_new("Rotation type");
   gtk_frame_set_shadow_type(GTK_FRAME(rotation_frame), GTK_SHADOW_ETCHED_IN);
-  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), rotation_frame,
-		     TRUE, TRUE, 5);
+  gtk_box_pack_start(GTK_BOX(main_vbox), rotation_frame, TRUE, TRUE, 5);
+  //gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), rotation_frame,
+  //		     TRUE, TRUE, 5);
 
   GtkWidget *rotation_vbox = gtk_vbox_new(TRUE, 2);
   gtk_container_add(GTK_CONTAINER(rotation_frame), rotation_vbox);
@@ -471,6 +542,13 @@ void options_button_clicked(GtkButton *button,
       pOpts->m_nRotationType = VNC_VIEWER_ROTATE_ALWAYS;
     else if (gtk_toggle_button_get_active(never_rotate_radio))
       pOpts->m_nRotationType = VNC_VIEWER_ROTATE_NEVER;
+
+    if (gtk_toggle_button_get_active(auto_encoding_radio))
+      pOpts->m_nEncodingType = VNC_VIEWER_ENCODING_AUTO;
+    else if (gtk_toggle_button_get_active(raw_encoding_radio))
+      pOpts->m_nEncodingType = VNC_VIEWER_ENCODING_RAW;
+    else if (gtk_toggle_button_get_active(hextile_encoding_radio))
+      pOpts->m_nEncodingType = VNC_VIEWER_ENCODING_HEXTILE;
   }
 
   gtk_widget_destroy(dialog);
@@ -559,6 +637,7 @@ static void vnc_connection_died(VncViewer *vnc, gpointer data) {
 			  mainview->default_opts.m_bShared,
 			  mainview->default_opts.m_nDepth,
 			  mainview->default_opts.m_nRotationType,
+			  mainview->default_opts.m_nEncodingType,
 			  mainview) ||
 	  !arg_lockdown)
 	break;
@@ -569,6 +648,7 @@ static void vnc_connection_died(VncViewer *vnc, gpointer data) {
 static gboolean make_connection(const gchar *display, gboolean shared,
 				VncViewerDepthType depth,
 				VncViewerRotationType rotationType,
+				VncViewerEncodingType encodingType,
 				MainView *mainview)
 {
   gint fd;
@@ -732,7 +812,7 @@ static gboolean make_connection(const gchar *display, gboolean shared,
   }
 
   // create VncViewer widget
-  mainview->vnc = VNC_VIEWER(vnc_viewer_new(fd, shared, depth));
+  mainview->vnc = VNC_VIEWER(vnc_viewer_new(fd, shared, depth, encodingType));
   mainview->vnc->mainview = (void *) mainview;
 
   switch (rotationType)
@@ -857,6 +937,7 @@ static void callback_connect(GtkBin *action, gpointer data)
   opts.m_bShared = mainview->default_opts.m_bShared;
   opts.m_nDepth = mainview->default_opts.m_nDepth;
   opts.m_nRotationType = mainview->default_opts.m_nRotationType;
+  opts.m_nEncodingType = mainview->default_opts.m_nEncodingType;
   // the "options" data is used by the Options dialog, it allows the opts
   // struct to be shared between the two dialogs
   g_object_set_data(G_OBJECT(dialog), "options", &opts);
@@ -925,12 +1006,14 @@ static void callback_connect(GtkBin *action, gpointer data)
       vnc_connection_closed(mainview);
 
     make_connection(display, opts.m_bShared, opts.m_nDepth,
-		    opts.m_nRotationType, mainview);
+		    opts.m_nRotationType, opts.m_nEncodingType,
+		    mainview);
 
     // preserve settings, even if the connection failed
     mainview->default_opts.m_bShared = opts.m_bShared;
     mainview->default_opts.m_nDepth = opts.m_nDepth;
     mainview->default_opts.m_nRotationType = opts.m_nRotationType;
+    mainview->default_opts.m_nEncodingType = opts.m_nEncodingType;
 
     nActive = gtk_combo_box_get_active(host_entry_combo);
     if (nActive != -1)
@@ -1515,6 +1598,20 @@ void scrolledwin_size_allocate(GtkWidget *widget,
   }
 }
 
+void scrolledwin2_size_request(GtkWidget *widget,
+			       GtkRequisition *requisition,
+			       gpointer userdata)
+{
+  // the height of the scrolled window isn't set properly by default,
+  // need to base it on size of the main vbox, which isn't calculated
+  // properly until we get to this point
+  GtkWidget *main_vbox = (GtkWidget *) userdata;
+  GtkRequisition requisition2;
+  gtk_widget_size_request(main_vbox, &requisition2);
+  // indicate to parent that we want to change the height
+  requisition->height = requisition2.height;
+}
+
 gint CompareHostsInList(const gchar *a, const gchar *b)
 {
   return strcmp(a, b);
@@ -1537,6 +1634,7 @@ static gboolean mainview_auto_connect(gpointer userdata)
 			    mainview->default_opts.m_bShared,
 			    mainview->default_opts.m_nDepth,
 			    mainview->default_opts.m_nRotationType,
+			    mainview->default_opts.m_nEncodingType,
 			    mainview))
 	{
 	  // move it to the beginning of m_hostList if the node already exists
@@ -1635,6 +1733,8 @@ static void create_menu_helper(MainView *mainview)
   gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(show_scrollbars_item),
 				 mainview->bShowScrollbars);
 
+  mainview->keyboard_item = gtk_menu_item_new_with_label(_("Keyboard"));
+
   mainview->help_item = gtk_menu_item_new_with_label(_("Help"));
   about_item = gtk_menu_item_new_with_label(_("About"));
 
@@ -1654,6 +1754,8 @@ static void create_menu_helper(MainView *mainview)
   gtk_menu_item_set_submenu(GTK_MENU_ITEM(mainview->file_item), file_menu);
   gtk_menu_item_set_submenu(GTK_MENU_ITEM(mainview->view_item), view_menu);
   gtk_menu_item_set_submenu(GTK_MENU_ITEM(toolbar_item), toolbar_menu);
+  gtk_menu_item_set_submenu(GTK_MENU_ITEM(mainview->keyboard_item),
+			    GTK_WIDGET(mainview->keys_menu));
   gtk_menu_item_set_submenu(GTK_MENU_ITEM(mainview->help_item), help_menu);
 
   // Attach the callback functions to the activate signal
@@ -1696,6 +1798,7 @@ void create_menu(void *mainvw)
   // Add menu items to right menus
   gtk_menu_append(main_menu, mainview->file_item);
   gtk_menu_append(main_menu, mainview->view_item);
+  gtk_menu_append(main_menu, mainview->keyboard_item);
   gtk_menu_append(main_menu, mainview->help_item);
   gtk_menu_append(main_menu, separator);
   gtk_menu_append(main_menu, mainview->quit_item);
