@@ -42,6 +42,7 @@
 #define MAEMO_VNCVIEWER_GCONF_SHARED MAEMO_VNCVIEWER_GCONF_PATH "/shared"
 #define MAEMO_VNCVIEWER_GCONF_DEPTH MAEMO_VNCVIEWER_GCONF_PATH "/depth"
 #define MAEMO_VNCVIEWER_GCONF_ROTATION MAEMO_VNCVIEWER_GCONF_PATH "/rotation"
+#define MAEMO_VNCVIEWER_GCONF_ENCODING MAEMO_VNCVIEWER_GCONF_PATH "/encoding"
 #define MAEMO_VNCVIEWER_GCONF_NORMAL_TOOLBAR MAEMO_VNCVIEWER_GCONF_PATH "/normal_toolbar"
 #define MAEMO_VNCVIEWER_GCONF_FULLSCREEN_TOOLBAR MAEMO_VNCVIEWER_GCONF_PATH "/full_screen_toolbar"
 #define MAEMO_VNCVIEWER_GCONF_SHOW_SCROLLBARS MAEMO_VNCVIEWER_GCONF_PATH "/show_scrollbars"
@@ -52,6 +53,7 @@ typedef struct options_settings
   gboolean m_bShared;
   VncViewerDepthType m_nDepth;
   VncViewerRotationType m_nRotationType;
+  VncViewerEncodingType m_nEncodingType;
 } options_settings;
 
 // Struct to include view's information
@@ -72,7 +74,7 @@ struct _MainView
   gboolean bMultiplePresses;
 
   // Items for menu
-  GtkWidget *file_item, *view_item, *help_item;
+  GtkWidget *file_item, *view_item, *keyboard_item, *help_item;
   GtkWidget *disconnect_item, *quit_item;
   GtkWidget *fullscreen_item;
   
