@@ -1,8 +1,8 @@
 /*
  * This file is part of vncviewer
  *
- * Copyright (C) 2005, 2006 Aaron Levinson.
- * Copyright (C) 2006, Detlef Schmicker
+ * Copyright (C) 2005-2007 Aaron Levinson.
+ * Copyright (C) 2006-2007 Detlef Schmicker
  *
  *
  * This software is free software; you can redistribute it and/or
@@ -253,13 +253,14 @@ static void vnc_viewer_init(VncViewer *self)
   self->mainview = 0;
 }
 
-GtkWidget *vnc_viewer_new(gint fd, gboolean shared, VncViewerDepthType depth)
+GtkWidget *vnc_viewer_new(gint fd, gboolean shared, VncViewerDepthType depth,
+			  VncViewerEncodingType encodingType)
 {
   VncViewer *vnc = gtk_type_new(vnc_viewer_get_type());
 
   if (!vnc)
     return NULL;
-  vnc_viewer_configure(vnc, fd, shared, depth);
+  vnc_viewer_configure(vnc, fd, shared, depth, encodingType);
 
   return GTK_WIDGET(vnc);
 }
@@ -339,17 +340,39 @@ GdkRgbCmap *cmap;
  * and bpp is the desired bpp
  */
 void vnc_viewer_configure(VncViewer *vnc, gint fd, gboolean shared,
-			  VncViewerDepthType depth)
+			  VncViewerDepthType depth,
+			  VncViewerEncodingType encodingType)
 {
-  guint32 encodings[] =
-  { 
-    rfbEncodingHextile,
-    rfbEncodingCoRRE,
-    rfbEncodingRRE,
-    rfbEncodingCopyRect, rfbEncodingRaw
-  };
+  guint32 encodings[5];
+  guint16 nEncodings;
+  switch (encodingType)
+  {
+  case VNC_VIEWER_ENCODING_AUTO:
+    encodings[0] = rfbEncodingHextile;
+    encodings[1] = rfbEncodingCoRRE;
+    encodings[2] = rfbEncodingRRE;
+    encodings[3] = rfbEncodingCopyRect;
+    encodings[4] = rfbEncodingRaw;
+    nEncodings = 5;
+    break;
+
+  case VNC_VIEWER_ENCODING_RAW:
+    encodings[0] = rfbEncodingCopyRect;
+    encodings[1] = rfbEncodingRaw;
+    nEncodings = 2;
+    break;
+
+  case VNC_VIEWER_ENCODING_HEXTILE:
+    encodings[0] = rfbEncodingHextile;
+    nEncodings = 1;
+    break;
+
+  default:
+    // shouldn't get here
+    nEncodings = 0;
+    break;
+  }
 
-  guint16 nEncodings = sizeof(encodings) / sizeof(encodings[0]);
   //guint32 localEncodings[] = { rfbEncodingCopyRect, rfbEncodingRaw };
   //guint nLocalEncodings = sizeof(localEncodings) / sizeof(localEncodings[0]);
 
@@ -412,7 +435,6 @@ void vnc_viewer_configure(VncViewer *vnc, gint fd, gboolean shared,
   g_assert(gdkvisual16->blue_shift==0);
   g_assert(gdkvisual16->type==GDK_VISUAL_TRUE_COLOR);
   
-
   switch (depth)
   {
   case VNC_VIEWER_DEPTH_8BIT:
@@ -421,7 +443,7 @@ void vnc_viewer_configure(VncViewer *vnc, gint fd, gboolean shared,
       // only change depth down
       vnc->pixel_format.bitsPerPixel = 8;
       vnc->pixel_format.depth = 8;
-      vnc->pixel_format.trueColour = TRUE;
+      vnc->pixel_format.trueColour = FALSE;
       vnc->pixel_format.redMax = 7;
       vnc->pixel_format.greenMax = 7;
       vnc->pixel_format.blueMax = 3;
@@ -490,8 +512,6 @@ void vnc_viewer_configure(VncViewer *vnc, gint fd, gboolean shared,
     */
   }
 
-
-
   vnc_viewer_queue_event(vnc, newSetPixelFormatMsg(&(vnc->pixel_format)));
 
   /*
@@ -648,60 +668,121 @@ static void vnc_viewer_framebuffer_update(VncViewer *vnc,
     {
     case rfbEncodingRaw:
       //g_message("raw encoding");
-      buffer = g_malloc(totPixels * 3);
-      for (i = 0; i < totPixels; i++)
+      if (bpp==8)
+      	buffer8=g_malloc(totPixels);
+      else
       {
-	guint32 pixel = 0;
-	switch (bpp)
-	{
-	case 8:
-	  pixel = desc->data.raw[i];
-	  break;
-
-	case 16:
-	  pixel = ((guint16 *)(desc->data.raw))[i];
-	  break;
-
-	case 32:
-	  pixel = ((guint32 *)(desc->data.raw))[i];
-	  break;
-
-	default:
-	  g_assert_not_reached();
-	}
-
 	if (!vnc->m_bRotateDisplay)
-	{
-	  buffer[3*i]   = (pixel >> fmt->redShift & fmt->redMax) * redScale;
-	  buffer[3*i+1] = (pixel >> fmt->greenShift & fmt->greenMax) *greenScale;
-	  buffer[3*i+2] = (pixel >> fmt->blueShift & fmt->blueMax) * blueScale;
-	}
+	  gdkimage=gdk_image_new(GDK_IMAGE_FASTEST, gdkvisual16, rWidth,
+				 rHeight);
+	else
+	  gdkimage=gdk_image_new(GDK_IMAGE_FASTEST, gdkvisual16, rHeight,
+				 rWidth);
+      }
+
+      guint ii = 0;
+      if (bpp == 8)
+      {
+	if (!vnc->m_bRotateDisplay)
+	  memcpy(buffer8, &(desc->data.raw), totPixels);
 	else
 	{
-	  j = rHeight * ((i % rWidth) + 1) - ((i / rWidth) + 1);
-	  buffer[3*j]   = (pixel >> fmt->redShift & fmt->redMax) * redScale;
-	  buffer[3*j+1] = (pixel >> fmt->greenShift & fmt->greenMax) *greenScale;
-	  buffer[3*j+2] = (pixel >> fmt->blueShift & fmt->blueMax) * blueScale;
+	  guint tmp;
+	  guint8 *tmpraw = (guint8 *) desc->data.raw;
+	  for (j = 0; j < rHeight; j++)
+	  {
+	    tmp = rHeight - 1 - j;
+	    for (i = 0; i < rWidth; i++)
+	    {
+	      buffer8[tmp] = *tmpraw;
+	      tmp += rHeight;
+	      tmpraw++;
+	    }
+	  }
 	}
       }
-	
-      if (!vnc->m_bRotateDisplay)
-      {
-	gdk_draw_rgb_image(vnc->offscreen, vnc->gc,
-			   desc->header.r.x, desc->header.r.y,
-			   rWidth, rHeight,
-			   GDK_RGB_DITHER_NORMAL, buffer, rWidth * 3);
-      }
       else
       {
-	gdk_draw_rgb_image(vnc->offscreen, vnc->gc,
-			   vnc->height - desc->header.r.y - rHeight,
-			   desc->header.r.x,
-			   rHeight, rWidth,
-			   GDK_RGB_DITHER_NORMAL, buffer, rHeight * 3);
+	if (!vnc->m_bRotateDisplay)
+	{
+	  if (gdkimage->bpl == (rWidth * 2))
+	  {
+	    // then can do a straight memcpy of the entire buffer
+	    //g_message("16-bit raw: able to do full memcpy");
+	    memcpy((guint8 *) gdkimage->mem, &(desc->data.raw), totPixels * 2);
+	  }
+	  else
+	  {
+	    // then need to do each row separately
+	    //g_message("16-bit raw: must do each row separately");
+	    for (j = 0; j < rHeight; j++)
+	    {
+	      memcpy(&((guint8 *) gdkimage->mem)[j * gdkimage->bpl],
+		     &(desc->data.raw)[ii], rWidth * 2);
+	      ii += rWidth * 2;
+	    }
+	  }
+	}
+	else
+	{
+	  // must use gdkimage->bpl, instead of rHeight, because
+	  // sometimes gdkimage->bpl is not equal to 2 * rHeight
+	  // (for instance, if rHeight is an odd number)
+	  guint16 *imgbegin = ((guint16 *) gdkimage->mem) + (rHeight - 1);
+	  guint16 *tmpimg;
+	  guint16 *tmpraw = (guint16 *) desc->data.raw;
+	  
+	  for (j = 0; j < rHeight; j++)
+	  {
+	    tmpimg = imgbegin - j;
+	    for (i = 0; i < rWidth; i++)
+	    {
+	      *tmpimg = *tmpraw;
+	      tmpimg = (guint16 *) (((guint8 *) tmpimg) + gdkimage->bpl);
+	      tmpraw++;
+	    }
+	  }
+	}
+      }
+
+      if (bpp == 8)
+      {
+	if (!vnc->m_bRotateDisplay)
+	{
+	  gdk_draw_indexed_image(vnc->offscreen, vnc->gc,
+				 desc->header.r.x, desc->header.r.y,
+				 rWidth, rHeight,
+				 GDK_RGB_DITHER_NONE, buffer8, rWidth, cmap);
+	}
+	else
+	{
+	  gdk_draw_indexed_image(vnc->offscreen, vnc->gc,
+				 vnc->height - desc->header.r.y - rHeight,
+				 desc->header.r.x,
+				 rHeight, rWidth,
+				 GDK_RGB_DITHER_NONE, buffer8, rHeight, cmap);
+	}
+	g_free(buffer8);
+	buffer8=NULL;
+      }
+      else
+      {
+	if (!vnc->m_bRotateDisplay)
+	{
+	  gdk_draw_image(vnc->offscreen, vnc->gc, gdkimage, 0, 0,
+			 desc->header.r.x, desc->header.r.y,
+			 rWidth, rHeight);
+	}
+	else
+	{
+	  gdk_draw_image(vnc->offscreen, vnc->gc, gdkimage, 0, 0,
+			 vnc->height - desc->header.r.y - rHeight,
+			 desc->header.r.x,
+			 rHeight, rWidth);
+	}
+	g_object_unref(gdkimage);
       }
 
-      g_free(buffer);
       break;
 
     case rfbEncodingCopyRect:
@@ -884,7 +965,6 @@ static void vnc_viewer_framebuffer_update(VncViewer *vnc,
     // May break in case this is changed in GDK
     case rfbEncodingHextile:
       //g_message("hextile encoding");
-      //Testing
       if (bpp==8)
       	buffer8=g_malloc(totPixels);
       else
@@ -922,12 +1002,19 @@ static void vnc_viewer_framebuffer_update(VncViewer *vnc,
 	    }
 	    else
 	    {
+	      guint beginpos = (xPos * rHeight) + (rHeight - yPos - 1);
+	      guint tmp;
+	      guint8 *tmpraw = (guint8 *) ht->data.raw;
 	      for (j = 0; j < ht->height; j++)
 	      {
+		tmp = beginpos - j;
 		for (i = 0; i < ht->width; i++)
 		{
-		  buffer8[((xPos + i) * rHeight) + (rHeight - yPos - j - 1)] =
-		    (ht->data.raw)[ii++];
+		  buffer8[tmp] = *tmpraw;
+		  tmp += rHeight;
+		  tmpraw++;
+		  //buffer8[((xPos + i) * rHeight) + (rHeight - yPos - j - 1)] =
+		  // (ht->data.raw)[ii++];
 		}
 	      }
 	    }
@@ -954,13 +1041,27 @@ static void vnc_viewer_framebuffer_update(VncViewer *vnc,
 	    }
 	    else
 	    {
+	      // must use gdkimage->bpl, instead of rHeight, because
+	      // sometimes gdkimage->bpl is not equal to 2 * rHeight
+	      // (for instance, if rHeight is an odd number)
+	      guint beginpos =
+		(xPos * gdkimage->bpl) + ((rHeight - yPos - 1) * 2);
+	      guint16 *imgbegin =
+		(guint16 *) (((guint8 *) gdkimage->mem) + beginpos);
+	      guint16 *tmpimg;
+	      guint16 *tmpraw = (guint16 *) ht->data.raw;
+
 	      for (j = 0; j < ht->height; j++)
 	      {
+		tmpimg = imgbegin - j;
 		for (i = 0; i < ht->width; i++)
 		{
-		  guint32 pixel = ((guint16 *)(ht->data.raw))[ii++];
-		  gdk_image_put_pixel(gdkimage, rHeight - yPos - j - 1,
-				      xPos + i, pixel);
+		  *tmpimg = *tmpraw;
+		  tmpimg = (guint16 *) (((guint8 *) tmpimg) + gdkimage->bpl);
+		  tmpraw++;
+		  //guint32 pixel = ((guint16 *)(ht->data.raw))[ii++];
+		  //gdk_image_put_pixel(gdkimage, rHeight - yPos - j - 1,
+		  //		      xPos + i, pixel);
 		}
 	      }
 	    }
@@ -1267,51 +1368,25 @@ static void vnc_viewer_framebuffer_update(VncViewer *vnc,
 	// g_message("xpos %d ypos %d",xPos,yPos);
       }
 
-      // gdkimage in the case of hextile 16 bit encoding, buffer8 in case of hextile 8 bit, 
-      // buffer in case of no hextile encoding
-      if (!gdkimage)
+      if (bpp == 8)
       {
-	if (!buffer8)
+	if (!vnc->m_bRotateDisplay)
 	{
-	  if (!vnc->m_bRotateDisplay)
-	  {
-	    gdk_draw_rgb_image(vnc->offscreen, vnc->gc,
-			       desc->header.r.x, desc->header.r.y,
-			       rWidth, rHeight,
-			       GDK_RGB_DITHER_NORMAL, buffer, rWidth * 3);
-	  }
-	  else
-	  {
-	    gdk_draw_rgb_image(vnc->offscreen, vnc->gc,
-			       vnc->height - desc->header.r.y - rHeight,
-			       desc->header.r.x,
-			       rHeight, rWidth,
-			       GDK_RGB_DITHER_NORMAL, buffer, rHeight * 3);
-	  }
-	  g_free(buffer);
+	  gdk_draw_indexed_image(vnc->offscreen, vnc->gc,
+				 desc->header.r.x, desc->header.r.y,
+				 rWidth, rHeight,
+				 GDK_RGB_DITHER_NONE, buffer8, rWidth, cmap);
 	}
 	else
 	{
-	  //g_message("draw attempt");
-	  if (!vnc->m_bRotateDisplay)
-	  {
-	    gdk_draw_indexed_image(vnc->offscreen, vnc->gc,
-				   desc->header.r.x, desc->header.r.y,
-				   rWidth, rHeight,
-				   GDK_RGB_DITHER_NONE, buffer8, rWidth, cmap);
-	  }
-	  else
-	  {
-	    gdk_draw_indexed_image(vnc->offscreen, vnc->gc,
-	    			   vnc->height - desc->header.r.y - rHeight,
-				   desc->header.r.x,
-				   rHeight, rWidth,
-				   GDK_RGB_DITHER_NONE, buffer8, rHeight, cmap);
-	  }
-	  g_free(buffer8);
-	  buffer8=NULL;
-	  //g_message("draw done");
+	  gdk_draw_indexed_image(vnc->offscreen, vnc->gc,
+				 vnc->height - desc->header.r.y - rHeight,
+				 desc->header.r.x,
+				 rHeight, rWidth,
+				 GDK_RGB_DITHER_NONE, buffer8, rHeight, cmap);
 	}
+	g_free(buffer8);
+	buffer8=NULL;
       }
       else
       {
